#include <cstdio>
#include <cstring>
#include <vector>
#include <list>

using namespace std;

struct channel
{
    enum chantype { undefined=0, sqr=1, rnd=2, tri=3, pcm=4 } type;
    unsigned char vol;     // 0..15 for any
    unsigned wavelen;      // 0..0x7FF for any
    unsigned char sample;  // or duty for sqr(2,4,8,12)
    unsigned age;          // Age of this row: most useful with pcm
    
    unsigned ins; // Filled later
};

class song
{
public:
    /* Parameters */
    unsigned rate;
    unsigned channelcount;
    /* Data */
    unsigned rownum;
    vector<channel> chanptr;
    vector<unsigned char> dpcmcache[256];

    song() : rate(0), channelcount(0), rownum(0), chanage(0)
    {
    }
    
private:
    /* Variables */
    unsigned chanage;
    
    void CheckRow()
    {
        unsigned oldrownum = rownum;
        rownum += chanage;
        if(chanptr.size() <= rownum*channelcount)
        {
            const unsigned increment = 1024;
            chanptr.resize(((rownum + increment) & ~(increment-1)) * channelcount);
        }
        
        for(unsigned tmp=0; tmp<chanage; ++tmp)
        {
               unsigned t2 = tmp+1;
               unsigned src  = oldrownum * channelcount;
               unsigned dest = src + t2 * channelcount;
            for(unsigned c=0; c<channelcount; ++c)
            {
                chanptr[dest+c] = chanptr[src+c];
                chanptr[dest+c].age = chanptr[src+c].age + t2;
            }
        }
        chanage=0;
    }
public:
    void AddAge(unsigned n)
    {
        chanage += n;
    }
    
    const channel &ChanPtr(unsigned row, unsigned chan) const
    {
        return chanptr[row * channelcount + chan];
    }

    channel &GetChan(unsigned c)
    {
        CheckRow();
        channel &tmp = chanptr[rownum * channelcount + c];
        tmp.age = 0;
        return tmp;
    }
};

#define CONTENTDEBUG 0
static song ReadSND(const char *fn)
{
    song song;
    FILE *fp = fopen(fn, "rb");
    
    if(!fp) { perror(fn); return song; }
    
    unsigned char Buf[64];
    fread(Buf, 16, 1, fp);
    if(Buf[0]!='S' || Buf[1]!='N' || Buf[2]!='D' || Buf[3]!=26)
    {
        fclose(fp);
        fprintf(stderr, "%s: Not a SND file\n", fn);
        return song;
    }
    
    int version = Buf[4];
    
    song.channelcount = Buf[5];
    song.rate         = Buf[6];
    
    if(!song.channelcount || song.channelcount>=0x20 || !song.rate)
    {
        fclose(fp);
        fprintf(stderr, "%s: Improper settings in file\n", fn);
        return song;
    }
    
    for(;;)
    {
        int c = fgetc(fp);
        if(c == EOF)break;
#if CONTENTDEBUG
        fprintf(stderr, "%02X  ", c);
#endif
        if(c == 0xFB)
        {
            int n = fgetc(fp), sl = fgetc(fp), sh = fgetc(fp);
            song.dpcmcache[n].resize(sl + sh*256);
            fread(&song.dpcmcache[n][0], 1, song.dpcmcache[n].size(), fp);
            continue;
        }
        if(c == 0xFE)
        {
            int c = fgetc(fp), t = fgetc(fp);
            
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            channel &tmp = song.GetChan(c);
            tmp.type = (channel::chantype)t;
            continue;
        }
        if(c == 0xFF)
        {
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            song.AddAge(1);
            continue;
        }
        if(c == 0xFD)
        {
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            song.AddAge(fgetc(fp));
            continue;
        }
        if(version == 3 &&
           ((c & 0xF0) == 0x20
         || (c & 0xF0) == 0x40
         || (c & 0xF0) == 0x80
         || (c & 0xF0) == 0xC0))
        {
            int duty = c>>4, vol=c&15;
            int c=fgetc(fp), wl=fgetc(fp), wh=fgetc(fp);
            
            channel &tmp = song.GetChan(c);
            tmp.wavelen = wl + wh*256;
            tmp.vol = vol;
            tmp.sample = duty;
            continue;
        }
        if((unsigned)c < song.channelcount)
        {
            int wl = fgetc(fp), wh = fgetc(fp), vol = fgetc(fp);
            
#if CONTENTDEBUG
            fprintf(stderr, "\b\b\b\b[%u,%u,%u]  ",c,vol,wl+wh*256);
#endif
            channel &tmp = song.GetChan(c);
            tmp.wavelen = wl + wh*256;
            
            if(tmp.type == channel::undefined)
            {
                int t = (c==0 || c==1) ? channel::sqr : (c==2 ? channel::tri : channel::rnd);
                if(t == channel::sqr)tmp.sample = 4;
                tmp.type = (channel::chantype)t;
            }

            if(version==1) { vol /= 16; tmp.wavelen = 110000 / (wl+wh*256); }
            if(version==1 && c>1) tmp.wavelen /= 2;
            
            if(tmp.type == channel::pcm)
            {
                tmp.sample = vol;
                tmp.vol = vol ? 6 : 0;
            }
            else
                tmp.vol = vol;
            continue;
        }
    }
    song.GetChan(0);
    fclose(fp);
    fprintf(stderr, "Got %u rows (at rate %u, %u seconds)\n",
        song.rownum, song.rate, (song.rownum + song.rate - 1) / song.rate);
    return song;
}

static bool IsTickPoint(const channel *prev, const channel *now)
{
    double wlendiff;
    if(!prev)return now;
    
    if(now->wavelen == 0 || prev->wavelen == 0)
        wlendiff = 0;
    else if(now->wavelen > prev->wavelen)
        wlendiff = now->wavelen / (double) prev->wavelen;
    else
        wlendiff = prev->wavelen / (double) now->wavelen;
    
    if(now->vol >= prev->vol+4
    || !prev->vol
    || (prev->vol > 1 && now->vol >= prev->vol*3)
    || (now->type != channel::rnd && wlendiff >= 1.057))
    {
        return true;
    }
    return false;
}

struct sampleitem
{
    int duty; /* Duty at this row */
    /* Frequency changes to the original */
    int wavelenadd;
    long wavelenmul;
    
    /* Volume changes to the original */
    int voladd;
    long volmul;
};
struct sample
{
    channel::chantype type;
    unsigned beginrow;
    unsigned basewavelen;
    unsigned char basevol;

	// FIXME: TM PITISI MUUTTAA NIIN, ETT ON 
	//   noin 7 elementin ALKUtree ja 7 elementin LOPPUtree.
	// Vain nit verrataan.
    list<sampleitem> tree;
    
    sample() : beginrow(0),basewavelen(0),basevol(0) {}
    
    enum comparetype { undefined=0, add, mul };
    /* Whether the wavelen is compared with multiplication or addition */
    comparetype wavelentype;
    /* Whether the volume is compared with multiplication or addition */
    comparetype voltype;

    /* For noise channels, duty doesn't count */
    /* For triangle channels, duty and volume don't count */
    
    const bool dutytest() const { return type == channel::sqr; }
    const bool volutest() const { return type != channel::tri; }
    const bool wavetest() const { return type != channel::sqr; }
};

static vector<sample> repository;

struct SampleMatchType
{
    unsigned length;
    bool waveaddmatch, wavemulmatch;
    bool voladdmatch, volmulmatch;
    
    SampleMatchType() : length(0),
      waveaddmatch(true), wavemulmatch(true),
      voladdmatch(true), volmulmatch(true) { }
    SampleMatchType(bool) : length(0) { }
};

// Test whether the "book" defines "data"
static SampleMatchType TestMatch(const sample &book, const sample &data)
{
    typedef list<sampleitem>::const_iterator itemit;
    
    // Different type samples can never match
    if(book.type != data.type)
    {
        return false;
    }
    
    const bool dutytest = book.dutytest();
    const bool volutest = book.volutest();
    const bool wavetest = book.wavetest();
    
    itemit bookit = book.tree.begin();
    itemit datait = data.tree.begin();
    
    // Note: data has all "types" undefined.
    // If the book has them defined, they master the comparison.
    
    SampleMatchType match;
    
    while(bookit != book.tree.end()
       && datait != data.tree.end())
    {
        if(dutytest && bookit->duty != datait->duty) return false;
        if(wavetest)
        {
            const bool addmatch = bookit->wavelenadd == datait->wavelenadd;
            const bool mulmatch = bookit->wavelenmul == datait->wavelenmul;
/*
            fprintf(stderr, "[Len %u, %u%s%u %ld%s%ld]\n", match.length,
                bookit->wavelenadd,addmatch?"==":"!=", datait->wavelenadd,
                bookit->wavelenmul,mulmatch?"==":"!=", datait->wavelenmul);
*/
            if(!addmatch) { if(book.wavelentype == sample::add)return false; match.waveaddmatch = false; if(!match.wavemulmatch)return false; }
            if(!mulmatch) { if(book.wavelentype == sample::mul)return false; match.wavemulmatch = false; if(!match.waveaddmatch)return false; }
        }
        if(volutest)
        {
            const bool addmatch = bookit->voladd == datait->voladd;
            const bool mulmatch = bookit->volmul == datait->volmul;
/*
            fprintf(stderr, "[Len %u, %u%s%u %ld%s%ld]\n", match.length,
                bookit->voladd,addmatch?"==":"!=", datait->voladd,
                bookit->volmul,mulmatch?"==":"!=", datait->volmul);
*/
            if(!addmatch) { if(book.voltype == sample::add)return false; match.voladdmatch = false; if(!match.volmulmatch)return false; }
            if(!mulmatch) { if(book.voltype == sample::mul)return false; match.volmulmatch = false; if(!match.voladdmatch)return false; }
        }
        ++bookit;
        ++datait;
        ++match.length;
        #if 0
        if(match.length >= 7)
        {
            while(bookit != book.tree.end()
               && datait != data.tree.end())
            {
                ++bookit;
                ++datait;
                ++match.length;
            }
            break;
        }
        #endif
    }
    // Found a match. {wave,vol}{add,mul}match could be used
    // now to tell which kind of matching we should be doing.
    return match;
}

static void DumpSample(const sample &ins)
{
    fprintf(stderr, "   Type %s\n",
        (ins.type==channel::sqr)?"SQR"
       :(ins.type==channel::tri)?"TRI"
       :(ins.type==channel::rnd)?"RND"
       :(ins.type==channel::pcm)?"PCM":"???");
    if(ins.volutest())
        fprintf(stderr, "   Base volume: %u\n"
                        "   Volume test type: %s\n",
            ins.basevol,
            ins.voltype==sample::undefined?"UNDEFINED"
           :ins.voltype==sample::add?"ADD"
           :ins.voltype==sample::mul?"MUL":"???");
    if(ins.wavetest())
        fprintf(stderr, "   Base wavelen: %u\n"
                        "   Wavelen test type: %s\n",
            ins.basewavelen,
            ins.wavelentype==sample::undefined?"UNDEFINED"
           :ins.wavelentype==sample::add?"ADD"
           :ins.wavelentype==sample::mul?"MUL":"???");
    fprintf(stderr, "   Length: %u samples\n", ins.tree.size());
    list<sampleitem>::const_iterator i;
    for(i=ins.tree.begin(); i!=ins.tree.end(); ++i)
    {
        fprintf(stderr, "     ");
        if(ins.dutytest())
            fprintf(stderr, " Duty %d ", i->duty);
        if(ins.volutest())
        {
            fprintf(stderr, " Volume");
            if(ins.voltype!=sample::mul)fprintf(stderr, " %+d", i->voladd);
            if(ins.voltype!=sample::add)fprintf(stderr, " *%ld", i->volmul);
            fprintf(stderr, " ");
        }
        if(ins.wavetest())
        {
            fprintf(stderr, " Wavelen");
            if(ins.wavelentype!=sample::mul)fprintf(stderr, " %+d", i->wavelenadd);
            if(ins.wavelentype!=sample::add)fprintf(stderr, " *%ld", i->wavelenmul);
            fprintf(stderr, " ");
        }
        fprintf(stderr, "\n");
    }
}

static unsigned Analyze(const sample &smp)
{
    /*
    fprintf(stderr, "--------------------\n");
    DumpSample(smp);
    */

    SampleMatchType bestmatch; unsigned bestmatchnum=0;
    for(unsigned a=0; a<repository.size(); ++a)
    {
        //fprintf(stderr, "COMPARING TO INS %u\n", a+1);
        SampleMatchType match = TestMatch(repository[a], smp);
        if(!match.length)continue;
        if(match.length > bestmatch.length)
        {
            bestmatch    = match;
            bestmatchnum = a;
        }
    }
    
/*
    if(!bestmatch.length)
        fprintf(stderr, "->No match!\n");
    else
    {
        fprintf(stderr, "->Best match:\n"
                        "  Length %u\n", bestmatch.length);
        if(smp.wavetest())
            fprintf(stderr, "  Wavelen match: %s,%s\n",
                bestmatch.waveaddmatch?"Add":"",
                bestmatch.wavemulmatch?"Mul":"");
        if(smp.volutest())
            fprintf(stderr, "  Volume match: %s,%s\n",
                bestmatch.voladdmatch?"Add":"",
                bestmatch.volmulmatch?"Mul":"");
    }
*/
    
    if(bestmatch.length)
    {
        sample &book = repository[bestmatchnum];
        sample newsample = book;
        bool changes = false;
        /* Upgrade the sample cache with possibly available matching type information */
        if(newsample.volutest() && newsample.voltype == sample::undefined)
        {
            if(bestmatch.voladdmatch && !bestmatch.volmulmatch)
            {
                changes = true;
                newsample.voltype = sample::add;
            }
            else if(bestmatch.volmulmatch && !bestmatch.voladdmatch)
            {
                changes = true;
                newsample.voltype = sample::mul;
            }
        }
        if(newsample.wavetest() && newsample.wavelentype == sample::undefined)
        {
            if(bestmatch.waveaddmatch && !bestmatch.wavemulmatch)
            {
                changes = true;
                newsample.wavelentype = sample::add;
            }
            else if(bestmatch.wavemulmatch && !bestmatch.waveaddmatch)
            {
                changes = true;
                newsample.wavelentype = sample::mul;
            }
        }
        if(smp.tree.size() > newsample.tree.size())
            newsample.tree = smp.tree;
        /*if(changes)
        {
            repository.push_back(newsample);
            return repository.size();
        }
        else*/
        {
            //fprintf(stderr, "UPDATED INSTRUMENT %u:\n", bestmatchnum+1);
            //DumpSample(newsample);
            book = newsample;
        }
        return bestmatchnum+1;
    }
    else
    {
        sample tmp = smp;
        repository.push_back(tmp);
        
        //fprintf(stderr, "CACHED INSTRUMENT %u:\n", repository.size());
        //DumpSample(tmp);
        
        /* None was found */
        return repository.size();
    }
}

static void AnalyzeSND(song &song)
{
    /* For DPCM channels the DPCM equipment only
     * takes care of instruments and they are rarely
     * tweaked, so we can take them just as raw */
     
    /* Often, it seems, samples start on one channel just a tick
     * before/after other channels. In these cases, it's considerable
     * to include that void tick in the sample.
     * Another way could be to move the sample backwards/forwards
     * for one tick measure...
     */

    vector<unsigned> lasttick(song.channelcount);
    for(unsigned c=0; c<song.channelcount; ++c)
        lasttick[c] = 0;
    vector<sample> cursample(song.channelcount);

    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const channel &tmp = song.ChanPtr(a, c);
            if(tmp.type == channel::pcm)continue;
            
            if(tmp.vol)
            {
                const channel *prev = a > 0 ? &song.ChanPtr(a-1, c) : NULL;
                
                sample &chansample = cursample[c];
                
                if(IsTickPoint(prev, &tmp))
                {
                    if(a != lasttick[c] + 1)
                    {
                        if(chansample.tree.size())
                        {
                            unsigned ins = Analyze(chansample);
                            song.chanptr[chansample.beginrow * song.channelcount + c].ins = ins;
                        }
                        /* Start of a new sample! */
                        chansample.type        = tmp.type;
                        chansample.beginrow    = a;
                        chansample.basewavelen = tmp.wavelen;
                        chansample.basevol     = tmp.vol;
                        chansample.tree.clear();
                    }
                    lasttick[c] = a;
                }

                if(chansample.basewavelen)
                {
                    sampleitem item;
                    item.duty       = tmp.sample;
                    item.wavelenadd = tmp.wavelen - chansample.basewavelen;
                    item.wavelenmul = tmp.wavelen * 2 / chansample.basewavelen;
                    item.voladd     = tmp.vol - chansample.basevol;
                    item.volmul     = tmp.vol * 2 / chansample.basevol;
                    
                    chansample.tree.push_back(item);
                }
            }
        }
    }
    
    for(unsigned a=0; a<repository.size(); ++a)
    {
        fprintf(stderr, "Instrument %u:\n", a+1);
        DumpSample(repository[a]);
    }
}

static void DumpSND(const song &song)
{
    vector<unsigned> lasttick(song.channelcount);
    for(unsigned c=0; c<song.channelcount; ++c)
        lasttick[c] = 0;

    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const channel &tmp = song.ChanPtr(a, c);
            char Buf[64];
            int colormode=0;
            
            sprintf(Buf, "v%Xs%dw%ua%u", tmp.vol, tmp.sample, tmp.wavelen, tmp.age);
            if(tmp.ins)
                sprintf(strchr(Buf,0), "(%u)", tmp.ins);
            
            sprintf(Buf+32, "%18s", Buf);
            if(tmp.vol)
            {
                const channel *prev = a > 0 ? &song.ChanPtr(a-1, c) : NULL;
                
                if(IsTickPoint(prev, &tmp))
                {
                    if(a != lasttick[c]+1)
                    {
                        fputs("\33[1;34m", stderr);
                        colormode=2;
                    }
                    else
                        colormode=1;
                    
                    lasttick[c] = a;
                }
                else
                    colormode=1;
            }
            else
            {
                colormode=0;
                fputs("\33[1;30m", stderr);
            }
            for(char *s=Buf+32; *s; ++s)
            {
                if(colormode==1)
                if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[1m", stderr);
                fputc(*s, stderr);
                if(colormode==1)
                if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[m", stderr);
            }
            fputs("\33[m", stderr);
        }
        fprintf(stderr, "\n");
    }
}

#include <sys/ioctl.h>
#include <stdlib.h>
#include "oss.c"
static char RandBuf[0x400];
static void BuildRand(void)
{
    for(unsigned a=0;a<0x400;)RandBuf[a++]=(rand()/4173)%2;
}
static int randnum(int ind)
{
    return RandBuf[ind];
}
#include "filters.c"
static unsigned GetInterPValue
   (const vector<unsigned char> &Buf, unsigned pos, unsigned div, unsigned speed)
{
#if 0 /* NO INTERPOLATION */
	speed=speed;
	return Buf[(pos/div) % Buf.size()];
#endif
#if 0 /* LINEAR INTERPOLATION */
	speed=speed;
    unsigned base = Buf[(pos        /div) % Buf.size()];
    unsigned next = Buf[((pos+div-1)/div) % Buf.size()];

    unsigned midpos = pos % div;

    return (next*midpos + base*(div-1-midpos)) / div;
#endif
#if 1 /* SLUGGISH INTERPOLATION [9=Cubic] */
	const FilterInfo &Filter = Filters[9];
	double scale_factor = max(1.0, speed / (double)div);
	double support      = max(0.5, scale_factor * Filter.support);
	if(support <= 0.5) { support = 0.5 + 1E-12; scale_factor = 1.0; }
	double *weight = (double *)alloca((sizeof(double)) * ((int)support*2+3));
	double (*const filterfun)(double) = Filter.function;
	const double center = pos / (double)div;
	int start = (int)(0.5 + center - support);
	int end   = (int)(0.5 + center + support);
	register double density = 0;
	int n = 0;
	for(int i = start; i < end; ++i)
		density += (weight[n++] = filterfun((i-center+0.5) / scale_factor) / scale_factor);
    if(density != 0.0)
        for(register int i=0; i<n; ++i)
            weight[i] /= density; /* normalize */
    register double result = 0.0;
    for(register unsigned i=0, k=start; i<n; ++i)
    	result += weight[i] * Buf[k++ % Buf.size()];
    return (int)result;
#endif
}

static void PlaySND(const song &song)
{
    vector<unsigned> waveptr(song.channelcount);

    int Bits=16, Stereo=0, Rate=44100;
    unsigned a;
    BuildRand();
    SetSoundSettings(1,&Bits,&Stereo,&Rate);
    
    for(a=0; a<song.rownum; ++a)
    {
        unsigned slen = Rate / song.rate;
        
        vector<short> Buf(slen);
        
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const channel &tmp = song.ChanPtr(a, c);
            if(!tmp.wavelen)continue;
            
            if(false
            //|| c==0
            //|| c==1
            //|| c==2
            //|| c==3
            //|| c==4
              )continue;
            
            for(unsigned p=0; p<slen; ++p)
            {
                signed char pha=0;
                int v;
                if(tmp.type == channel::tri)
                {
                    static const signed char wave[32] =
                    {0,1,2,3,4,5,6,7,7,6,5,4,3,2,1,0,
                    0,-1,-2,-3,-4,-5,-6,-7,-7,-6,-5,-4,-3,-2,-1,0};

                    pha = wave[(waveptr[c]/100000) & 31];
                }
                else if(tmp.type == channel::sqr)
                {
                    pha = (((waveptr[c]/100000) & 15) < tmp.sample) ? 15 : -15;
                }
                else if(tmp.type == channel::rnd)
                {
                    pha = randnum((waveptr[c]/400000) % 0x400) ? 5 : -5;
                }
                else if(tmp.type == channel::pcm)
                {
                    if(tmp.age==0)waveptr[c]=0;
                    const vector<unsigned char> &smp = song.dpcmcache[tmp.sample];
                    if(smp.size())
                    {
                    	pha = GetInterPValue(smp, waveptr[c], 100000, 4000000 / tmp.wavelen);
                        pha = (pha - 0x80) * 3;
                    }
                }
                
                v = pha * tmp.vol;
                Buf[p] += v*30;
                waveptr[c] += 4000000 / tmp.wavelen;
            }
        }
        fwrite(&Buf[0], sizeof(Buf[0]), Buf.size(), stdout);
    }
}

int main(int argc, const char *const *argv)
{
    if(argc != 2 || !strcmp(argv[1], "--help") || !strcmp(argv[1], "/?"))
    {
        puts("Simple SNDv3 player - Copyright (C) 1992,2002 Bisqwit http://bisqwit.iki.fi/)\n"
             "Usage:\n"
             "   snd <filename.snd>\n"
            );
        return 1;
    }
    
    song song = ReadSND(argv[1]);
    
    /* TODO: Handle the snd (dump it somewhere or something) */
    
    AnalyzeSND(song);
    DumpSND(song);
    PlaySND(song);
    
    return 0;
}
